SQLAlchemy 0.5 Documentation - module sqlalchemy.ext.declarative

# SQLAlchemy 0.5 Documentation

Multiple Pages | One Page
Version: 0.5.0rc3 Last Updated: 11/07/08 13:55:35

### module sqlalchemy.ext.declarative

A simple declarative layer for SQLAlchemy ORM.

SQLAlchemy object-relational configuration involves the usage of Table, mapper(), and class objects to define the three areas of configuration. declarative moves these three types of configuration underneath the individual mapped class. Regular SQLAlchemy schema and ORM constructs are used in most cases:

    from sqlalchemy.ext.declarative import declarative_base

    Base = declarative_base()

    class SomeClass(Base):
        __tablename__ = 'some_table'
        id = Column('id', Integer, primary_key=True)
        name =  Column('name', String(50))

Above, the `declarative_base` callable produces a new base class from which all mapped classes inherit from. When the class definition is completed, a new `Table` and `mapper()` have been generated, accessible via the `__table__` and `__mapper__` attributes on the `SomeClass` class.

You may omit the names from the Column definitions. Declarative will fill them in for you:

    class SomeClass(Base):
        __tablename__ = 'some_table'
        id = Column(Integer, primary_key=True)
        name = Column(String(50))

Attributes may be added to the class after its construction, and they will be added to the underlying `Table` and `mapper()` definitions as appropriate:

    SomeClass.data = Column('data', Unicode)
    SomeClass.related = relation(RelatedInfo)

Classes which are mapped explicitly using `mapper()` can interact freely with declarative classes.

The `declarative_base` base class contains a `MetaData` object where newly defined `Table` objects are collected. This is accessed via the `metadata` class level accessor, so to create tables we can say:

    engine = create_engine('sqlite://')
    Base.metadata.create_all(engine)

The `Engine` created above may also be directly associated with the declarative base class using the `engine` keyword argument, where it will be associated with the underlying `MetaData` object and allow SQL operations involving that metadata and its tables to make use of that engine automatically:

    Base = declarative_base(engine=create_engine('sqlite://'))

Or, as `MetaData` allows, at any time using the `bind` attribute:

    Base.metadata.bind = create_engine('sqlite://')

The `declarative_base` can also receive a pre-created `MetaData` object, which allows a declarative setup to be associated with an already existing traditional collection of `Table` objects:

    mymetadata = MetaData()
    Base = declarative_base(metadata=mymetadata)

Relations to other classes are done in the usual way, with the added feature that the class specified to `relation()` may be a string name. The "class registry" associated with `Base` is used at mapper compilation time to resolve the name into the actual class object, which is expected to have been defined once the mapper configuration is used:

    class User(Base):
        __tablename__ = 'users'

        id = Column(Integer, primary_key=True)
        name = Column(String(50))
        addresses = relation("Address", backref="user")

    class Address(Base):
        __tablename__ = 'addresses'

        id = Column(Integer, primary_key=True)
        email = Column(String(50))
        user_id = Column(Integer, ForeignKey('users.id'))

Column constructs, since they are just that, are immediately usable, as below where we define a primary join condition on the `Address` class using them:

    class Address(Base)
        __tablename__ = 'addresses'

        id = Column(Integer, primary_key=True)
        email = Column(String(50))
        user_id = Column(Integer, ForeignKey('users.id'))
        user = relation(User, primaryjoin=user_id == User.id)

In addition to the main argument for `relation`, other arguments which depend upon the columns present on an as-yet undefined class may also be specified as strings. These strings are evaluated as Python expressions. The full namespace available within this evaluation includes all classes mapped for this declarative base, as well as the contents of the `sqlalchemy` package, including expression functions like `desc` and `func`:

    class User(Base):
        # ....
        addresses = relation("Address", order_by="desc(Address.email)",
            primaryjoin="Address.user_id==User.id")

As an alternative to string-based attributes, attributes may also be defined after all classes have been created. Just add them to the target class after the fact:

    User.addresses = relation(Address, primaryjoin=Address.user_id == User.id)

Synonyms are one area where `declarative` needs to slightly change the usual SQLAlchemy configurational syntax. To define a getter/setter which proxies to an underlying attribute, use `synonym` with the `descriptor` argument:

    class MyClass(Base):
        __tablename__ = 'sometable'

        _attr = Column('attr', String)

        def _get_attr(self):
            return self._some_attr
        def _set_attr(self, attr)
            self._some_attr = attr
        attr = synonym('_attr', descriptor=property(_get_attr, _set_attr))

The above synonym is then usable as an instance attribute as well as a class-level expression construct:

    x = MyClass()
    x.attr = "some value"
    session.query(MyClass).filter(MyClass.attr == 'some other value').all()

The synonym_for decorator can accomplish the same task:

    class MyClass(Base):
        __tablename__ = 'sometable'

        _attr = Column('attr', String)

        @synonym_for('_attr')
        @property
        def attr(self):
            return self._some_attr

Similarly, comparable_using is a front end for the comparable_property ORM function:

    class MyClass(Base):
        __tablename__ = 'sometable'

        name = Column('name', String)

        @comparable_using(MyUpperCaseComparator)
        @property
        def uc_name(self):
            return self.name.upper()

As an alternative to `__tablename__`, a direct `Table` construct may be used. The `Column` objects, which in this case require their names, will be added to the mapping just like a regular mapping to a table:

    class MyClass(Base):
        __table__ = Table('my_table', Base.metadata,
            Column('id', Integer, primary_key=True),
            Column('name', String(50))
        )

Other table-based attributes include `__table_args__`, which is either a dictionary as in:

    class MyClass(Base)
        __tablename__ = 'sometable'
        __table_args__ = {'mysql_engine':'InnoDB'}

or a dictionary-containing tuple in the form `(arg1, arg2, ..., {kwarg1:value, ...})`, as in:

    class MyClass(Base)
        __tablename__ = 'sometable'
        __table_args__ = (ForeignKeyConstraint(['id'], ['remote_table.id']), {'autoload':True})

Mapper arguments are specified using the `__mapper_args__` class variable. Note that the column objects declared on the class are immediately usable, as in this joined-table inheritance example:

    class Person(Base):
        __tablename__ = 'people'
        id = Column(Integer, primary_key=True)
        discriminator = Column(String(50))
        __mapper_args__ = {'polymorphic_on': discriminator}

    class Engineer(Person):
        __tablename__ = 'engineers'
        __mapper_args__ = {'polymorphic_identity': 'engineer'}
        id = Column(Integer, ForeignKey('people.id'), primary_key=True)
        primary_language = Column(String(50))

For single-table inheritance, the `__tablename__` and `__table__` class variables are optional on a class when the class inherits from another mapped class.

As a convenience feature, the `declarative_base()` sets a default constructor on classes which takes keyword arguments, and assigns them to the named attributes:

    e = Engineer(primary_language='python')

Note that `declarative` has no integration built in with sessions, and is only intended as an optional syntax for the regular usage of mappers and Table objects. A typical application setup using `scoped_session` might look like:

    engine = create_engine('postgres://scott:tiger@localhost/test')
    Session = scoped_session(sessionmaker(autocommit=False,
                                          autoflush=False,
                                          bind=engine))
    Base = declarative_base()

Mapped instances then make usage of `Session` in the usual way.

### Module Functions

def comparable_using(comparator_factory)

Decorator, allow a Python @property to be used in query criteria.

A decorator front end to comparable_property(), passes through the comparator_factory and the function being decorated:

    @comparable_using(MyComparatorType)
    @property
    def prop(self):
        return 'special sauce'

The regular `comparable_property()` is also usable directly in a declarative setting and may be convenient for read/write properties:

    prop = comparable_property(MyComparatorType)

def declarative_base(bind=None, metadata=None, mapper=None, cls=, name='Base', constructor=, metaclass=, engine=None)

Construct a base class for declarative class definitions.

The new base class will be given a metaclass that invokes instrument_declarative() upon each subclass definition, and routes later Column- and Mapper-related attribute assignments made on the class into Table and Mapper assignments. See the declarative module documentation for examples.

bind
  An optional Connectable, will be assigned to the metadata.bind. The engine keyword argument is a deprecated synonym for bind.
metadata
  An optional MetaData instance. All Tables implicitly declared by subclasses of the base will share this MetaData. A MetaData instance will be create if none is provided. The MetaData instance will be available via the metadata attribute of the generated declarative base class.
mapper
  An optional callable, defaults to sqlalchemy.orm.mapper. Will be used to map subclasses to their Tables.
cls
  Defaults to object. A type to use as the base for the generated declarative base class. May be a type or tuple of types.
name
  Defaults to 'Base', Python's internal display name for the generated class. Customizing this is not required, but can improve clarity in tracebacks and debugging.
constructor
  Defaults to declarative._declarative_constructor, an __init__ implementation that assigns **kwargs for declared fields and relations to an instance. If None is supplied, no __init__ will be installed and construction will fall back to cls.__init__ with normal Python semantics.
metaclass
  Defaults to DeclarativeMeta. A metaclass or __metaclass__ compatible callable to use as the meta type of the generated declarative base class.

def instrument_declarative(cls, registry, metadata)

Given a class, configure the class declaratively, using the given registry (any dictionary) and MetaData object. This operation does not assume any kind of class hierarchy.

def synonym_for(name, map_column=False)

Decorator, make a Python @property a query synonym for a column.

A decorator version of synonym(). The function being decorated is the 'descriptor', otherwise passes its arguments through to synonym():

    @synonym_for('col')
    @property
    def prop(self):
        return 'special sauce'

The regular `synonym()` is also usable directly in a declarative setting and may be convenient for read/write properties:

    prop = synonym('col', descriptor=property(_read_prop, _write_prop))